﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SearchAndSorting
{
    /// <summary>
    /// Сервисный класс, содержащий различные методы сортировки
    /// </summary>
    /// <typeparam name="T">тип элементов сортируемого массива</typeparam>
    public class SortService<T> where T : IComparable<T>
    {
        public delegate void SortMethod(T[] arr); 

        /// <summary>
        /// Вызывает рекурсивную процедуру быстрой сортировки
        /// </summary>
        /// <param name="arr">сортируемый массив</param>
        public static void QuickSort(T[] arr)
        {
            int n = arr.Length;
            quickSort(arr, 0, n - 1);
        }
        /// <summary>
        /// Небольшая по размеру процедура содержит три
        /// вложенных цикла while, три оператора if и рекурсивные
        /// вызовы. Для таких процедур задание инвариантов и
        /// доказательство корректности облегчает отладку. 
        /// Предусловие: (left <= right)
        /// Постусловие: массив arr отсортирован по возрастанию
        /// </summary>
        /// <param name="arr">сортируемый массив</param>
        /// <param name="left">начальный индекс сортируемой части 
        /// массива arr</param>
        /// <param name="right">конечный индекс сортируемой части 
        /// массива arr</param>       
        protected static void quickSort(T[] arr, int left, int right)
        {
            T item = arr[(left + right) / 2];
            int l = left;
            int r = right;
            T temp;
            /* 
            Введем три непересекающихся множества:
                 S1: {arr(i), left <= i =< l-1}  
                 S2: {arr(i), l <= i =< r}  
                 S1: {arr(i), r+1 <= i =< right}
                 Введем следующие логические условия, 
                 играющие роль инвариантов циклов нашей программы:
                 P1: объединение S1, S2, S3 = arr
                 P2: (S1(i) <= item) Для всех элементов S1
                 P3: (S3(i) >= item) Для всех элементов S3
                 P4: item - элемент arr
                 Нетрудно видеть, что все условия становятся 
                 истинными после завершения инициализатора цикла.
                 Для пустых множеств S1 и S3 условия P2 и P3
                 считаются истинными по определению.
                 Inv = P1 & P2 & P3 & P4 = true 
                ********/
            while (l <= r) //непроверенная зона S2 не пуста 
            {
                while (arr[l].CompareTo(item) == -1) l++;
                /*
                 * расширяем зону S1, сохраняя истинность P2 и INV
                 * Цикл завершается, когда arr[l] >= item
                 * Цикл всегда завершится, поскольку вначале гарантируется,
                 * что элемент item находится в интервале [l,r],
                 * а после первого обмена зона S3 не пуста, 
                 * что гарантирует завершаемость.                
                 * */
                while (arr[r].CompareTo(item) == 1) r--;
                /*
                * Цикл всегда завершится и arr[r] < item 
                * */
                if (l <= r)
                {
                    /*
                     * Оба элемента с индексами l и r находятся
                     * в непроверенной зоне. Можно выполнить обмен,
                     * сохраняя истинность инварианта INV 
                     * и расширяя зоны S1 и S3
                     * */
                    temp = arr[l]; arr[l] = arr[r]; arr[r] = temp;
                    l++; r--;
                }
                /*
                 * Иначе непроверенная зона пуста. Цикл завершится! 
                 * */
            }
            /* 
             * По завершению цикла:
             * Инвариант INV остается истинным
             * S2 - пусто
             * r левее l
             * Зона между l и r содержит только элементы item 
            */

            if (left < r)
                quickSort(arr, left, r);//Рекурсивный вызов для S1             
            if (right > l)
                quickSort(arr, l, right);//Рекурсивный вызов для S3            
        }
       
        //HeapSort - SortTree - PyramidSort
        /// <summary>
        /// Рекурсивная поцедура просеивания элемента по древу (пирамиде)
        /// Основная компонента построения пирамиды для массива
        /// и построения отсортированного массива по пирамиде
        /// </summary>
        /// <param name="i"></param>
        static void Seed(T[] arr, int i, int Heap_Size)
        {
            //просеивание элемента с индексом i,
            //нарушающего свойство пирамидальности массива
            int l = 2 * i + 1, r = l + 1;   //индексы левого и правого потомка
            T temp;
            if (l > Heap_Size - 1) return;  //это лист пирамиды
            int cand = i;
            if (arr[i].CompareTo(arr[l]) ==-1) cand = l;
            if ((r < Heap_Size) && (arr[cand].CompareTo(arr[r]) == -1) )
                cand = r;
            if (cand != i) //обмен
            {
                temp = arr[i]; arr[i] = arr[cand]; arr[cand] = temp;
                Seed(arr, cand, Heap_Size);     //Просеивание вниз
            }
        }
        /// <summary>
        /// Преобразует массив arr  в пирамиду
        /// Массив является пирамидой, если для каждого элемента массива,
        /// не являющегося листом пирамиды, его значение больше или равно
        /// значения двух его потомков
        /// </summary>
        static void MakePyramid(T[] arr, int Heap_Size)
        {
            int m = (arr.Length - 2) / 2;
            //m- индекс последнего элемента, имеющего потомков
            //элементы с индексами, большими m, являются листьями пирамиды
            for (int k = m; k >= 0; k--)
                Seed(arr, k, Heap_Size);
        }
        /// <summary>
        /// Пирамидальная сортировка
        /// Вначале по массиву строится пирамида 
        /// затем по пирамиде строится отсортированнный массив
        /// </summary>
        public static void PyramidSort(T[] arr)
        {
            int Heap_Size = arr.Length;
            T temp;
            MakePyramid(arr, Heap_Size);           
            for (int i = 0; i < arr.Length - 1; i++)
            {
                temp = arr[0]; arr[0] = arr[Heap_Size - 1];
                arr[Heap_Size - 1] = temp; Heap_Size--;
                Seed(arr, 0, Heap_Size);
            }
        }

       
       /// <summary>
       /// Сортировка слиянием
       /// Вызывает рекурсивную процедуру SortM,
       /// передавая ей границы сортируемого массива.
       /// Сортируемый массив arr задается
       /// соответствующим полем класса.
       public static void SortMerge(T[] arr)
       {
           int n = arr.Length;
           T[] temp = new T[n];
           SortM(arr, 0, n - 1, temp);
       }
       /// <summary>
       /// Рекурсивная процедура сортировки слиянием
       /// </summary>
       /// <param name="start">начало сортируемой части массива</param>
       /// <param name="finish">конец сортируемой части массива</param>
       /// <param name="temp">массив для хранения результатов слияния</param>
       static void SortM(T[] arr, int start, int finish, T[] temp)
       {
           if (start < finish)
           {
               int mid = (start + finish) / 2;
               SortM(arr, start, mid, temp);
               SortM(arr, mid + 1, finish, temp);
               Merge(arr, start, mid, finish, temp);
           }
       }
       /// <summary>
       /// Слияние отсортированных частей масссива в массив temp
       /// После слияния массив переписывается в исходный массив
       /// </summary>
       /// <param name="start">начало первой части</param>
       /// <param name="mid">конец первой части - начало второй</param>
       /// <param name="finish">конец второй части</param>
       /// <param name="temp">временный массив с результатами слияния</param>
       /// 
       static void Merge(T[] arr, int start, int mid, int finish, T[] temp)
       {
           int topl = start, topr = mid + 1, topt = start;
           //Слияние, пока не завершится одна из частей массива
           while ((topl <= mid) && (topr <= finish))
           {
               if (arr[topl].CompareTo(arr[topr]) < 1)
                   temp[topt++] = arr[topl++];
               else
                   temp[topt++] = arr[topr++];
           }
           //дописывание остатка незавершенной части массива
           while (topl <= mid)
               temp[topt++] = arr[topl++];
           while (topr <= finish)
               temp[topt++] = arr[topr++];
           //переливаем элементы в исходный массив
           for (int i = start; i <= finish; i++)
               arr[i] = temp[i];
       }

        public static long HowLong(SortMethod sort, int count,
           T[] arr)
        {
            T[] sortarr = new T[arr.Length];
            DateTime start, finish;
            start = DateTime.Now;

            for (int i = 0; i < count; i++)
            {
                Array.Copy(arr, sortarr, arr.Length);
                sort(sortarr);
            }

            finish = DateTime.Now;
            return finish.Ticks - start.Ticks;

        }
        /// <summary>
        /// Сортировка Шелла
        /// </summary>
        /// <param name="arr"></param>
        public static void ShellSort(T[] arr)
        {
            int n = arr.Length, m = (int)Math.Log(n, 2);
            //Создать последовательность чисел Шелла - h
            int[] h = new int[m];
            int t = 2;
            for (int i = 0; i < m; i++)
            {
                h[i] = t - 1; t *= 2;
            }
            //Внешний цикл по последовательности чисел h
            //Внутренний по числу групп
            for (int i = m-1; i >=0; i--)
                for (int j = 0; j < h[i]; j++)
                    //SortInsert(arr, h[i], j);
                    SortInsertOLD(arr, h[i], j);
        }
        /// <summary>
        /// Сортировка вставкой группы элементов массива arr,
        /// начинающейся с элемента, заданного индексом ind,
        /// и отстоящих друг от друга на расстоянии h
        /// </summary>
        /// <param name="arr">массив</param>
        /// <param name="h">расстояние между элементами в группе</param>
        /// <param name="ind">индекс начального элемента группы</param>
        private static void SortInsert(T[] arr, int h, int ind)
        {
            int n = arr.Length;
            int k = ind + h, i = 0;
            T item;
            while (k < n)
            {
                //вставка arr[k] на место
                i = k - h;
                while (i >= ind)
                {
                    if (arr[i+h].CompareTo(arr[i]) == -1)
                    {
                        item = arr[i]; arr[i] = arr[i + h];
                        arr[i + h] = item; i -= h;
                    }
                    else break; 
                }
                k += h;
            }

        }

        /// <summary>
        /// Сортировка вставкой группы элементов массива arr,
        /// начинающейся с элемента, заданного индексом ind,
        /// и отстоящих друг от друга на расстоянии h
        /// </summary>
        /// <param name="arr">массив</param>
        /// <param name="h">расстояние между элементами в группе</param>
        /// <param name="ind">индекс начального элемента группы</param>
        private static void SortInsertOLD(T[] arr, int h, int ind)
        {
            int n = arr.Length;
            int k = ind + h, i = 0;
            T item;
            while (k < n)
            {
                //вставка arr[k] на место
                i = k - h;
                while (i >= ind)
                {
                    if (arr[i + h].CompareTo(arr[i]) == -1)
                    {
                        item = arr[i]; arr[i] = arr[i + h];
                        arr[i + h] = item; i -= h;
                    }
                    else break;
                }
                k += h;
            }

        }

        /// <summary>
        /// Сортировать массив, 
        /// Предусловие: Элементы массива принадлежат двум видам,
        /// заданным в массиве Spacimen
        /// </summary>
        /// <param name="ar">сортируемый массив</param>
        /// <param name="Spacimen">массив представителей</param>
        public static void SortTwoKinds(T[] ar, T[] Spacimen)
        {
            int start = 0, finish = ar.Length - 1;
            T val1 = Spacimen[0], val2 = Spacimen[1];
            while (start < finish)
            {
                while ((start < finish) && (ar[start].CompareTo(val1) == 0))
                    start++;
                while ((start < finish) && (ar[finish].CompareTo(val2) == 0))
                    finish--;
                //обмен
                T temp = ar[start]; ar[start] = ar[finish];
                ar[finish] = temp;
                start++; finish--;
            }
        }
        /// <summary>
        /// Пузырьковая сортировка
        /// </summary>
        /// <param name="arr">сортируемый массив</param>
        public static void BubbleSort(T[] arr)
        {
            int n = arr.Length;
            bool done = false;
            int i = 0;
            T temp;
            while (!done)
            {   
                done = true;
                for (int j = n-1; j > i; j--)
                {
                    if (arr[j].CompareTo(arr[j - 1]) == -1)
                    {
                        temp = arr[j];
                        arr[j] = arr[j - 1];
                        arr[j - 1] = temp;
                        done = false;                         
                    }
                }
            

               
            }
        }
        /// <summary>
        /// Сортировка вставкой
        /// </summary>
        /// <param name="arr"></param>
        public static void InsertSort(T[] arr)
        {
            int n = arr.Length;
            for (int i = 1; i < n; i++)
            {
               //ставим на место элемент с индексом i
                T temp = arr[i];
                int j = i - 1;
                while ((j >= 0) && (arr[j].CompareTo(temp) == 1) )
                {
                    arr[j + 1] = arr[j];
                    j--;
                }
                arr[j+1] = temp;
            }
        }
    }
         
}
